

//
// Functions that support the auto-expand of names in the code textarea
//


var OFF_SCREEN_LEFT = '-10000px';


// Cache for previously-encountered prefix-expansion lists (cleared when ruletype or transaction type changes)
var Prefix_Expansion_Cache = new Array();
function clearExpansionCache()
{
   Prefix_Expansion_Cache = new Array();
}


// Defines legal expansions for all field names
var Field_Expansions;
function loadFieldExpansions()
{
  Field_Expansions = new Array();
  for (var f in FieldLabelMap)
  {
    if (FieldLabelMap[f] instanceof Function) continue;
    Field_Expansions.push(FieldLabelMap[f]);
  }
  Field_Expansions.sort(caseInsensitiveSort);
}



// Defines legal expansions for macros for the current rule type
var Macro_Expansions;
function loadMacroExpansions()
{
  Macro_Expansions = new Array();
  var ruleTypeId = document.ruleForm.ruleTypeId.value;
  var ruleTypeChar = ruleTypeLabelMap[ruleTypeId] ? ruleTypeLabelMap[ruleTypeId].substring(0,1) : "?";
  var sel = document.getElementById("alertType");
  var alertType = sel.options[sel.selectedIndex].value;
  if (ruleTypeId  &&  ruleTypeId.replace(/^\s*|\s*$/g,"")!='')  // Trim whitespace
  {
    for (var i=0; i<actionCode.length; i++)
    {
      var rt = actionCode[i].split(" ")[0];
      if (ruleTypeId == rt)
      {
          var code = actionCode[i].substring(actionCode[i].indexOf(" ") + 1);
          if (ruleTypeChar!="Q")
              ;
          else if (("*"+alertType).indexOf(code.substring(0,1)) >= 0)
              code = code.substring(1);
          else
              continue;
        Macro_Expansions.push( code + " : " + actionLabel[i] );
      }
    }
  }
  Macro_Expansions.sort(caseInsensitiveSort);
}

function caseInsensitiveSort(x, y)
{
    try {
       var a = String(x).toUpperCase();
       var b = String(y).toUpperCase();
       if (a > b)
           return 1;
       if (a < b)
           return -1
       return 0;
    } catch(err) {
        // alert('a=' + a + ', b=' + b + ': ' + err.description);
        return 0;
    }
}

// Defines overview list of expansions for this rule type (segments, macros, ruleNumbers, uservars)
var Overview_Expansions;
function loadOverviewExpansions()
{
  Overview_Expansions = new Array();

  // Insert labels for non-uservars segments
  var compSegs = getComponentSegmentIds();
  var uVarSegs = new Object();
  for (var s=0; s<compSegs.length; s++) {
    var lab = SegmentLabelMap[compSegs[s]];
    if (lab) {
      Overview_Expansions.push(lab);
      uVarSegs[compSegs[s]] = s;
    }
  }

  // Each uservar segment requires segments for all of its keys
  for (var uvarSid in UserVarSegmentKeys) {
      var found = true;
      for (var ii=0; ii<UserVarSegmentKeys[uvarSid].length; ii++) {
          var keySid = UserVarSegmentKeys[uvarSid][ii];
          if (!uVarSegs[keySid]) {
              found = false;
              break;
          }
      }
      if (found)
      {
          if (SegmentLabelMap[uvarSid]) // happens when no fields
              Overview_Expansions.push(SegmentLabelMap[uvarSid]);
      }
  }

  Overview_Expansions.push( "': Common Message Field Values" );
  Overview_Expansions.push( "%: Rule Macros" );
  Overview_Expansions.push( "#: Eligible Prerequisite Rules" );
  Overview_Expansions.sort(caseInsensitiveSort);
}


// Defines legal prereq rules from the _depends array plus Q rules (if ARD or uservar rule)
var Prereq_Expansions;
function loadPrereqExpansions()
{
  Prereq_Expansions = new Array();
  for (var p in preRequisiteLabelMap)
  {
    if (preRequisiteLabelMap[p] instanceof Function) continue;
    Prereq_Expansions.push(preRequisiteLabelMap[p]);
  }

  // This sorts the rules by ruleType/ruleName/ruleNumber
  // Prereq_Expansions.sort(prereqComparator);
  // This sorts the rules by ruleNumber
  Prereq_Expansions.sort();
}


// Undo previous change to code
var theUndoStack = new Array();
var theUndoCursor = -1;
var theUndoStackLimit = 256;
function undo()
{
  if (theUndoCursor >= 0)
  {
    var textArea = document.getElementById("code");
    if (theUndoStack[theUndoCursor] == textArea.value)
       theUndoCursor--;
    if (theUndoCursor >= 0)
    {
       textArea.value = theUndoStack[theUndoCursor--];
       theUndoStack.push(textArea.value);
       if (theUndoStack.length > theUndoStackLimit)
       {
          theUndoStack.shift();
          theUndoCursor--;
       }
    }
  }
  showUndoStack();
}
// Note action for possible later undoing
function noteForUndo()
{
  var textArea = document.getElementById("code");
  var lastOne = theUndoStack.length - 1;
  if (lastOne < 0 || textArea.value != theUndoStack[lastOne])
  {
    theUndoStack.push(textArea.value);
    if (lastOne > theUndoStackLimit)
      theUndoStack.shift();
    theUndoCursor = theUndoStack.length - 1;
  }
  //if (confirm("Click OK to Loop Forever!")) for (var i=0; true; i++) { i--; }
  showUndoStack();
}
// Take undoStack snapshot on mouse click
function noteMouseClick(evt)
{
  if (evt.type == 'click' || evt.type == 'contextmenu')
    noteForUndo();
}
// Show stack state
function showUndoStack()
{
  document.getElementById("undoStackSize").innerHTML = theUndoStack.length;
  return;
  var stackElementSizes = new Array();
  for (var i=0; i<theUndoStack.length; i++) stackElementSizes.push(theUndoStack[i].length);
  document.getElementById("stackElementSizes").innerHTML = stackElementSizes.join(",");
}



// Tracks state of cntl and alt keys (down or not-down)
NoteKey_Cntl_Key_Down = false;
NoteKey_Alt_Key_Down = false;
NoteKey_Shift_Key_Down = false;
Ignore_Shift_Space = false; // S0844689
function noteKey( keyEvent )
{
  keyEvent = (keyEvent) ? keyEvent : window.event;
  if (keyEvent.type == 'keydown')
  {
    if (keyEvent.keyCode == NoteKey_Cntl_Key_Code)
      NoteKey_Cntl_Key_Down = true;
//    else if (keyEvent.keyCode == NoteKey_Alt_Key_Code)
//      NoteKey_Alt_Key_Down = true;
    else if (keyEvent.keyCode == NoteKey_Shift_Key_Code)
      NoteKey_Shift_Key_Down = true;
    if (NoteKey_Cntl_Key_Down && keyEvent.keyCode==NoteKey_Z_Key_Code)
      undo();
    else
      noteForUndo();
  }
  else if (keyEvent.type == 'keyup')
  {
    if (keyEvent.keyCode == NoteKey_Cntl_Key_Code)
      NoteKey_Cntl_Key_Down = false;
//    else if (keyEvent.keyCode == NoteKey_Alt_Key_Code)
//      NoteKey_Alt_Key_Down = false;
    else if (keyEvent.keyCode == NoteKey_Shift_Key_Code)
      NoteKey_Shift_Key_Down = false;
  }

  return keyEvent.keyCode;
}
function setIgnoreShiftSpace(trueFalse) { // S0844689
    Ignore_Shift_Space = trueFalse;       // S0844689
}                                         // S0844689

NoteKey_Backspace_Key_Code = 8;
NoteKey_Shift_Key_Code = 16;
NoteKey_Cntl_Key_Code = 17;
NoteKey_Alt_Key_Code = 18;
NoteKey_Space_Key_Code = 32;
NoteKey_Right_Key_Code = 39;
NoteKey_Left_Key_Code = 37;
NoteKey_Delete_Key_Code = 46;
NoteKey_Quote_Key_Code = 222;
NoteKey_Pound_Key_Code = 51;
NoteKey_Percent_Key_Code = 53;
NoteKey_Z_Key_Code = 90;
//NoteKey_Ampersand_Key_Code = 55;
NoteKey_Underscore_Key_Code = (document.selection) ? 189 : 109; // IE vs FF



// Returns html for the expansion selector
function getExpansionSelectorHtml( startIndex, endIndex, textTag, prefix, divId, caretPos )
{
  // Different prefixes search different expansions
  var expansions = getExpansionsList( prefix, textTag, caretPos );
  //var eligibleUserVarTypes = document.getElementById("eligibleUserVarTypes").value;

  // Create a selector for the possibilities
  var listHTML = new Array("<pre class='famSelect'>");
  for (var i=startIndex; i<endIndex; i++)
  {
    var textToInsert = expansions[i].substring(prefix.length).split(':')[0].replace(/^\s*|\s*$/g,"");
    textToInsert = textToInsert.replace(/"/g,"\\\"").replace(/'/g,"\\x27"); //"
    listHTML.push( "<div style='border:1 solid #faf8ed;'");
    listHTML.push( " onmouseup='insertText(document.getElementById(\"" + textTag.id + "\"),\"" + textToInsert + "\");removeSelector();'" );
    listHTML.push( " onmouseover='this.style.backgroundColor=\"#cfdeeb\";this.style.border=\"1 solid #719db9\";'" );
    listHTML.push( " onmouseout ='this.style.backgroundColor=\"#faf8ed\";this.style.border=\"1 solid #faf8ed\";'" );
    listHTML.push( ">");
    listHTML.push( escapeXml(expansions[i].replace(/KEY;/g,"KEY:")) ); // For lookup lists
    listHTML.push( "</div>");
  }
  listHTML.push( "</pre>");
  return listHTML.join('');
}




// Hides the expansions selector when there is a mouseup
function handleMouseUp(e)
{
  var e = new CommonEvent(e);
  if (e.target.id == 'code'  ||  e.target.id == 'expansionSelectorDiv')
    removeSelector();
}

// Hides expansion selector and removes the listener
function removeSelector()
{
  removeListener("mouseup", handleMouseUp);
  document.getElementById("expansionSelectorDiv").style.left = OFF_SCREEN_LEFT;
}


// Expands the chars preceding the caret when requested
function expand( keyEvent, textTag, expansionSelectorDivId )
{
  var keyCode = noteKey(keyEvent);
  if (keyCode != NoteKey_Cntl_Key_Code  &&  keyCode != NoteKey_Shift_Key_Code  &&  keyCode != NoteKey_Alt_Key_Code)
  {
    // Enable the save button after any char (other textareas require focus change to re-enable)
    enableSaveButton( keyEvent );

    // If a char that could be part of the prefix was pressed and the selector is showing, recalc prefix and redisplay selector
    var expansionSelectorDiv = document.getElementById(expansionSelectorDivId);
    var caretPos = getCaretPos(textTag);
    //log('legal=' + keyCode + isLegalPrefixCode(keyCode) + ' hid=' + (expansionSelectorDiv.style.left == OFF_SCREEN_LEFT));
    if (isLegalPrefixCode(keyCode)  &&  expansionSelectorDiv.style.left != OFF_SCREEN_LEFT)
    {
      var prefix = getPrefix( textTag, caretPos );
      var possibilities = getExpansions( prefix.toLowerCase(), textTag, caretPos );
      if (possibilities)
        expansionSelectorDiv.innerHTML = getExpansionSelectorHtml(possibilities[0], possibilities[1], textTag, prefix, expansionSelectorDivId, caretPos);
      else
      {
        expansionSelectorDiv.style.left = OFF_SCREEN_LEFT;
        removeListener("mouseup", handleMouseUp);
      }
      //log('islegal&visible: ' + prefix + " " + possibilities);
    }

    // Else if cntl/alt/shift-space was pressed, expand the prefix and (re)display the selector
    else if ((NoteKey_Cntl_Key_Down || NoteKey_Alt_Key_Down || (NoteKey_Shift_Key_Down && !Ignore_Shift_Space))  &&  keyCode==NoteKey_Space_Key_Code) // S0844689
    {
      // alert("NoteKey_Cntl_Key_Down="+NoteKey_Cntl_Key_Down + " NoteKey_Alt_Key_Down="+NoteKey_Alt_Key_Down + " NoteKey_Shift_Key_Down="+NoteKey_Shift_Key_Down + " Ignore_Shift_Space="+Ignore_Shift_Space); // S0844689
      // Get prefix
      if (document.selection && (textTag.value.substring(caretPos-1,caretPos) == " ")) // Remove blank (IE only)
      {
        var newLineList = textTag.value.substring(0,caretPos).match(/\n/gm);
        newLineList = (newLineList) ? newLineList : new Array();
        setSelectionRange(textTag, caretPos-1-newLineList.length, caretPos-newLineList.length);
        replaceSelection( textTag, "");
        caretPos--;
      }
      var prefix = getPrefix(textTag, caretPos);

      // Identify possible prefix expansions
      var expansions = getExpansionsList( prefix, textTag, caretPos );
      var possibilities = getExpansions( prefix.toLowerCase(), textTag, caretPos );
      if (possibilities)
      {
        var startIndex = possibilities[0];
        var endIndex = possibilities[1];
        var commonPrefixLength = possibilities[2];

        // If only one possible expansion of a non-OverviewExpansion, fill it in and hide the selector
        if (startIndex == endIndex-1  &&  expansions != Overview_Expansions)
        {
          insertText( textTag, expansions[startIndex].substring(prefix.length).split(':')[0] );
          expansionSelectorDiv.style.left = OFF_SCREEN_LEFT;
          removeListener("mouseup", handleMouseUp);
        }

        // Else expand prefix as much as possible and display a selector for expansions
        else
        {
          // Set new prefix from additional common prefix chars
          if (commonPrefixLength > prefix.length)
          {
            insertText( textTag, expansions[startIndex].substring(prefix.length, commonPrefixLength) );
            prefix += expansions[ startIndex ].substring(prefix.length, commonPrefixLength).toLowerCase();
          }

          // If only one possible expansion, we must be in Overview_Expansions, so switch to a non-OverviewExpansion
          if (startIndex == endIndex-1  &&  expansions == Overview_Expansions)
          {
            expansions = getExpansionsList( prefix, textTag, caretPos );
            possibilities = getExpansions( prefix.toLowerCase(), textTag, caretPos );
            // If nothing to expand here, hide the selector and return
            if (!possibilities || possibilities.length==0)
            {
              expansionSelectorDiv.style.left = OFF_SCREEN_LEFT;
              removeListener("mouseup", handleMouseUp);
              return
            }
            startIndex = possibilities[0];
            endIndex = possibilities[1];
            commonPrefixLength = possibilities[2];
          }

          // Create a selector of expansions of the prefix on left or right based on chars since last newline
          var fontWidth = getFontWidthHeight(textTag)[0];
          var newLinePos = textTag.value.substring(0,caretPos).lastIndexOf("\n");
          var indentInEms = caretPos - newLinePos;
          var indentInPixels = indentInEms * fontWidth;
          var windowWidthInPoints = (window.innerWidth) ? window.innerWidth : document.body.offsetWidth;
          var windowWidthInPixels = windowWidthInPoints * 96 / 72;
          var divWidthInPixels = expansionSelectorDiv.offsetWidth;
          var textWidthInPixels = parseInt(textTag.style.width);
          if (indentInPixels < textWidthInPixels*.5)
            expansionSelectorDiv.style.marginLeft = (textWidthInPixels-divWidthInPixels-fontWidth) + "px";
          else
            expansionSelectorDiv.style.marginLeft = '2em';


          expansionSelectorDiv.innerHTML = getExpansionSelectorHtml(startIndex, endIndex, textTag, prefix, expansionSelectorDivId, caretPos);
          expansionSelectorDiv.style.left = '0px';

          // A click anywhere closes the selector
          addListener("mouseup", handleMouseUp);
        }
      }
    }

    // Else hide the selector
    else
    {
      expansionSelectorDiv.style.left = OFF_SCREEN_LEFT;
    }
  }
}



// Returns actual size of current font in pixels as [width,height]
function getFontWidthHeight(env) {
  env= env || document.body;
  var tmp= document.createElement('div');
  var atts= {fontSize:'1em',padding:'0',position:'absolute',lineHeight:'1',visibility:'hidden'};
  for(var p in atts){
    tmp.style[p]= atts[p];
  }
  tmp.appendChild(document.createTextNode('M'));
  env.appendChild(tmp);
  var widthHeight= [tmp.offsetWidth,tmp.offsetHeight];
  env.removeChild(tmp);
  return widthHeight;
}



// Returns an array identifying possible expansions for the prefix
// as follows:  [0]=startIndex, [1]=endIndex, [2]=commonPrefixLength
function getExpansions( prefix, textTag, caretPos )
{
  // We present different expansions for different prefixes
  var expansions = getExpansionsList( prefix, textTag, caretPos );
  //log('getExpansions: prefix='+prefix + " expansions=" + expansions);

  // Check the cache first
  var possibilities;
  if (prefix.substring(0,1)!="'" && prefix.substring(0,1)!="%") // cannot use cache for strings or macros
    possibilities = Prefix_Expansion_Cache[ prefix ];
  if (! possibilities)
  {
    var startIndex;
    var endIndex = expansions.length;
    for (startIndex=0; startIndex < expansions.length; startIndex++)
    {
      // On first prefix match, note it then search for the last match
      if (expansions[startIndex].substring(0,prefix.length).toLowerCase() == prefix)
      {
        for (endIndex=startIndex+1; endIndex < expansions.length; endIndex++)
        {
          if (expansions[endIndex].substring(0,prefix.length).toLowerCase() != prefix)
            break;
        }
        break;
      }
    }

    // Determine the longest matching prefix in the slice defined by startIndex/endIndex
    if (startIndex < endIndex)
    {
      var firstMatch = expansions[startIndex].split(":")[0];
      var lastMatch = expansions[endIndex-1].split(":")[0];
      var len;
      for (len=prefix.length; len<Math.min(firstMatch.length, lastMatch.length); len++)
      {
        if (firstMatch.substring(len,len+1) != lastMatch.substring(len,len+1))
          break;
      }
      // Remember what we found in the cache
      possibilities = new Array(startIndex, endIndex, len);
      Prefix_Expansion_Cache[ prefix ] = possibilities;
    }
  }
  return possibilities;
}



// Gets characters preceding the caret that are allowed to form the prefix.
// A prefix is a SAS name (alphanum+underscore), SAS macro (% plus SAS name),
// a hash name (includes dot in the name),
// a rule number (# plus digits), or a common constant (' after IN)
function getPrefix(textTag, caretPos)
{
  var lim = Math.max(0,caretPos-48);
  var i;
  for ( i=caretPos-1; i>=lim; --i)
  {
    var chr = textTag.value.substring(i,i+1).toLowerCase();
    // Continue building prefix if next char is part of legal SAS name
    if (Legal_SAS_Name_Chars_With_Dot.indexOf(chr) > -1)
      continue;
    // Prefix is complete on encountering % (start of a macro) or # (start of rule ID) or ' (start of constant)
    if ("%#'".indexOf(chr) > -1)
      break;
    else
      i++;
    break;
  }
  return textTag.value.substring(i,caretPos);
}
Legal_SAS_Name_Chars_With_Dot = ".abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890_";


// Returns true/false based on whether the keyCode is legal in a prefix or not
function isLegalPrefixCode(keyCode)
{
  if (NoteKey_Shift_Key_Down)
    return keyCode == NoteKey_Pound_Key_Code      ||
           keyCode == NoteKey_Percent_Key_Code    ||
           keyCode == NoteKey_Underscore_Key_Code ||
           Legal_SAS_Name_Chars_With_Dot.indexOf( String.fromCharCode(keyCode) ) > -1;
  else
    return keyCode == NoteKey_Quote_Key_Code      ||
           keyCode == NoteKey_Right_Key_Code      ||
           keyCode == NoteKey_Left_Key_Code       ||
           keyCode == NoteKey_Backspace_Key_Code  ||
           keyCode == NoteKey_Delete_Key_Code     ||
           keyCode == NoteKey_Underscore_Key_Code ||
           Legal_SAS_Name_Chars_With_Dot.indexOf( String.fromCharCode(keyCode) ) > -1;
}




// Inserts text in a textarea tag
function insertText(textTag, textToAdd)
{
//  textToAdd = textToAdd..replace/^\s*|\s*$/g,'');
  textToAdd = textToAdd.replace(/KEY;/g,'KEY:'); // For lookup lists
  if (document.selection)   // IE
  {
    textTag.focus();
    var range = document.selection.createRange();
    range.text = textToAdd;

    // Move cursor to end of inserted text (though recommended, this does not work)
    //range.moveStart('character', -textToAdd.length);
    //range.select();

    // If we inserted text ending with an insertion sequence (_ or # and a blank followed by close paren or comma), move back to the insertion point
    var ip = textToAdd.search(/(_|#) (\)|,)/);
    if (ip>=0) {
      range.moveEnd('character', ip-textToAdd.length+1);
      range.select();
    }
  }

  else if (textTag.selectionStart || textTag.selectionStart == '0')   // FF
  {
    var startPos = textTag.selectionStart;
    var endPos = textTag.selectionEnd;
    textTag.value = textTag.value.substring(0, startPos) + textToAdd + textTag.value.substring(endPos);
    // Move cursor to end of inserted text
    textTag.setSelectionRange(startPos+textToAdd.length,startPos+textToAdd.length);
    // @TODO: Move back to insertion point if inserted text ended with an insertion sequence
  }

  else     // Neither IE nor Firefox
  {
    textTag.value += textToAdd;
  }
  textTag.focus();
}



// Returns the list of expansions to check based on the length and beginning char of the prefix
// This expression allows comments anywhere within the IN(...) expression but run prohibitively slowly if the IN(...) expression is long
//var Constant_Prefix_RegExp = new RegExp(/(\b[0-9a-z_]+\b)(\s+in\s*\(\s*('[^']+'\s*(\/\*([^*]|[\r\n]|(\*+([^*\/]|[\r\n])))*\*+\/)?\s*,\s*(\/\*([^*]|[\r\n]|(\*+([^*\/]|[\r\n])))*\*+\/)?\s*)*|\s*=\s*|\s*~=\s*|\s+ne\s*|\s+eq\s*)'[^']*$/i);
// This expression runs very fast but does not allow comments within the IN() at all
//var Constant_Prefix_RegExp = new RegExp(/(\b[0-9a-z_]+\b)(\s+in\s*\(\s*('[^']+'\s*,\s*)*|\s*=\s*|\s*~=\s*|\s+ne\s*|\s+eq\s*)'[^']*$/i);
// This expression runs as fast as the one above and allows comments within the IN() only before the comma (not after)
var Constant_Prefix_RegExp = new RegExp(/(\b[0-9a-z_]+\b)(\s+in\s*\(\s*('[^']+'\s*(\/\*([^*]|[\r\n]|(\*+([^*\/]|[\r\n])))*\*+\/)?\s*,\s*)*|\s*=\s*|\s*~=\s*|\s+ne\s*|\s+eq\s*)'[^']*$/i); // '
function getExpansionsList( prefix, textTag, caretPos )
{
  var prefixLen = prefix.length;
  var chr = (prefixLen > 0) ? prefix.substring(0,1) : "";
  if (chr == "%")
    expansions = Macro_Expansions;
  else if (chr == "#")
    expansions = Prereq_Expansions;
  else if (chr == "'") {
    var matches = Constant_Prefix_RegExp.exec(textTag.value.substring(0, caretPos));
    if (matches == null)
      expansions = new Array();
    else {
      var fieldName = matches[1].toLowerCase();
      var type = CommonFieldValueTypes[fieldName];
      if (type)
        expansions = CommonFieldValues[type];
      else
        expansions = CommonFieldValues[fieldName];
      //log('getExpansionLists: prefix=' + prefix + ' caretPos=' + caretPos + ' matches=' + matches.join('+') + ' expansions=' + expansions);
    }
    if (!expansions)
      expansions = new Array();
  }
  else if (prefixLen <= 2)
    expansions = Overview_Expansions;
  else
    expansions = Field_Expansions;
  return expansions;
}


